@@ -64,6 +64,7 @@ extern "C" {
 #include "srl_iterator.h"
 #include "srl_decoder.h"
 #include "srl_reader_misc.h"
+#define WANT_SRL_RDR_TRACE /* allow use of SRL_RDR_TRACE */
 #include "srl_reader_error.h"
 #include "srl_reader_varint.h"
 #include "srl_reader_decompress.h"
@@ -144,8 +145,10 @@ extern "C" {
 #   define SRL_ITER_REPORT_STACK_STATE(iter)
 #endif
 
+// note that it's different from SRL_RDR_DONE()
+#define SRL_RDR_DONE_(buf) ((buf).pos > (buf).end)
 #define SRL_ITER_ASSERT_EOF(iter, msg) STMT_START {                                 \
-    if (expect_false(SRL_RDR_DONE((iter)->pbuf))) {                                 \
+    if (expect_false(SRL_RDR_DONE_((iter)->buf))) {                                 \
         SRL_RDR_ERROR_EOF((iter)->pbuf, (msg));                                     \
     }                                                                               \
 } STMT_END
@@ -482,7 +485,7 @@ read_again:
     iter->buf.pos++;                                                                            \
                                                                                                 \
     /* No code which decrease step, next or stack's counters should be added here.              \
-     * Otherwise the counters will be decreased twicer for tags like REFN, ALIAS, etc. */       \
+     * Otherwise the counters will be decreased twicer for tags like REFN, WEAKEN , etc. */     \
                                                                                                 \
     switch (tag & 0xE0) {                                                                       \
         case 0x0: /* POS_0 .. NEG_1 */                                                          \
@@ -526,7 +529,6 @@ read_again:
                     break;                                                                      \
                                                                                                 \
                 case SRL_HDR_REFN:                                                              \
-                case SRL_HDR_ALIAS:                                                             \
                 case SRL_HDR_WEAKEN:                                                            \
                     goto read_again;                                                            \
                                                                                                 \
@@ -543,6 +545,7 @@ read_again:
                                                                                                 \
                 case SRL_HDR_COPY:                                                              \
                 case SRL_HDR_REFP:                                                              \
+                case SRL_HDR_ALIAS:                                                             \
                     srl_skip_varint(aTHX_ iter->pbuf);                                          \
                     break;                                                                      \
                                                                                                 \
@@ -585,51 +588,43 @@ srl_iterator_step_in(pTHX_ srl_iterator_t *iter, UV n)
         SRL_ITER_REPORT_TAG(iter, tag);
         iter->buf.pos++;
 
-        switch (tag) {
-            case SRL_HDR_REFP:
-                /* store offset for REFP tag. Will be used in srl_iterator_wrap_stack() */
-                stack_ptr->prev_depth = orig_pos - iter->buf.body_pos; // almost same sa SRL_RDR_BODY_POS_OFS
+        if (tag == SRL_HDR_REFP || tag == SRL_HDR_ALIAS) {
+            /* store offset for REFP tag. Will be used in srl_iterator_wrap_stack() */
+            stack_ptr->prev_depth = orig_pos - iter->buf.body_pos; // almost same sa SRL_RDR_BODY_POS_OFS
 
-                offset = srl_read_varint_uv_offset(aTHX_ iter->pbuf, " while reading REFP tag");
-                iter->buf.pos = iter->buf.body_pos + offset;
-                DEBUG_ASSERT_RDR_SANE(iter->pbuf);
+            offset = srl_read_varint_uv_offset(aTHX_ iter->pbuf, " while reading REFP tag");
+            iter->buf.pos = iter->buf.body_pos + offset;
+            DEBUG_ASSERT_RDR_SANE(iter->pbuf);
 
-                tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;
-                SRL_ITER_TRACE_WITH_POSITION("tag SRL_HDR_REFP points to tag SRL_HDR_%s",
-                                             SRL_TAG_NAME(tag));
+            tag = *iter->buf.pos & ~SRL_HDR_TRACK_FLAG;
+            SRL_ITER_TRACE_WITH_POSITION("tag SRL_HDR_REFP points to tag SRL_HDR_%s",
+                                         SRL_TAG_NAME(tag));
 
-                switch (tag) {
-                    case SRL_HDR_HASH:
-                    case SRL_HDR_ARRAY:
-                        goto step_in;
+            switch (tag) {
+                case SRL_HDR_HASH:
+                case SRL_HDR_ARRAY:
+                    break;
 
-                    case SRL_HDR_OBJECTV:
-                    case SRL_HDR_OBJECTV_FREEZE:
-                    case SRL_HDR_OBJECT:
-                    case SRL_HDR_OBJECT_FREEZE:
-                        croak("not implemented OBJECT");
+                case SRL_HDR_OBJECTV:
+                case SRL_HDR_OBJECTV_FREEZE:
+                case SRL_HDR_OBJECT:
+                case SRL_HDR_OBJECT_FREEZE:
+                    croak("not implemented OBJECT");
 
-                    CASE_SRL_HDR_HASHREF:
-                    CASE_SRL_HDR_ARRAYREF:
-                        croak("not implemente HASHREF ARRAYREF");
+                CASE_SRL_HDR_HASHREF:
+                CASE_SRL_HDR_ARRAYREF:
+                    croak("not implemente HASHREF ARRAYREF");
 
-                    default:
-                        goto restore_pos_and_step_in;
-                }
-                break;
-
-            case SRL_HDR_COPY:
-            case SRL_HDR_ALIAS:
-                croak("not implemented COPY/ALIAS");
-
-            default:
-restore_pos_and_step_in:
-                stack_ptr->prev_depth = 0;
-                iter->buf.pos = orig_pos;
-                break;
+                default:
+                    stack_ptr->prev_depth = 0;
+                    iter->buf.pos = orig_pos;
+                    break;
+            }
+        } else {
+            stack_ptr->prev_depth = 0;
+            iter->buf.pos = orig_pos;
         }
 
-step_in:
         srl_iterator_step_internal(iter, stack_ptr);
     }
 
